/*
Copyright 1989, 1998  The Open Group

Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation.

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of The Open Group shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from The Open Group.
 *
 * Author:  Keith Packard, MIT X Consortium
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <X11/Xos.h>
#include <X11/X.h>
#include <X11/Xmd.h>
#include <X11/Xdmcp.h>

#ifdef HAVE_LIBBSD
#include <bsd/stdlib.h> /* for arc4random_buf() */
#endif

#ifndef HAVE_ARC4RANDOM_BUF
static void
getbits (long data, unsigned char *dst)
{
    dst[0] = (data      ) & 0xff;
    dst[1] = (data >>  8) & 0xff;
    dst[2] = (data >> 16) & 0xff;
    dst[3] = (data >> 24) & 0xff;
}
#endif

#define Time_t time_t

#include <stdlib.h>

#if defined(HAVE_LRAND48) && defined(HAVE_SRAND48)
#define srandom srand48
#define random lrand48
#endif
#ifdef WIN32
#include <process.h>
#define srandom srand
#define random rand
#define getpid(x) _getpid(x)
#endif

#ifndef HAVE_ARC4RANDOM_BUF

static void
insecure_getrandom_buf (unsigned char *auth, int len)
{
    long    lowbits, highbits;

    srandom ((int)getpid() ^ time((Time_t *)0));
    lowbits = random ();
    highbits = random ();
    getbits (lowbits, auth);
    getbits (highbits, auth + 4);
}

#ifndef _GNU_SOURCE
#define _GNU_SOURCE // needed on SLE11 for O_CLOEXEC
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/random.h>
#include <sys/syscall.h>

int getentropy_urandom(void *buffer, size_t length)
{
	int random_fd = -1;
	ssize_t res = -1;
	size_t filled = 0;

	if( length > 256 )
	{
		errno = EIO;
		return -1;
	}

	random_fd = open("/dev/urandom", O_RDONLY | O_CLOEXEC);

	if( random_fd == -1 )
	{
		return -1;
	}

	while( filled < length )
	{
		res = read(random_fd, (char*)buffer + filled, length - filled);

		if( res == -1 )
		{
			// shouldn't actually happen acc. to man(4) random,
			// but you never know
			if( errno == EINTR )
				continue;

			close(random_fd);
			return -1;
		}
		else if( res == 0 )
		{
			// no more bytes available? should not happen
			errno = EIO;
			close(random_fd);
			return -1;
		}

		filled += res;
	}

	close(random_fd);
	return 0;
}

int getentropy_getrandom(void *buffer, size_t length)
{
	int res;
	size_t filled = 0;

	if( length > 256 )
	{
		errno = EIO;
		return -1;
	}

	while( filled < length )
	{
#ifdef SYS_getrandom
		/*
		 * glibc does not contain a syscall wrapper for this in older
		 * versions
		 */
		res = syscall(SYS_getrandom, (char*)buffer + filled, length - filled, 0);
#else
#	warning no getrandom
		errno = ENOSYS;
		return -1;
#endif // SYS_getrandom

		if( res == -1 )
		{
			if( errno == EINTR )
				continue;

			return -1;
		}
		else if( res == 0 )
		{
			// no more bytes available? should not happen
			errno = EIO;
			return -1;
		}

		filled += res;
	}

	return 0;
}

int getentropy_emulate(void *buffer, size_t length)
{
	/*
	 * check at runtime whether we have a getrandom system call available,
	 * otherwise fall back to urandom approach. autoconf check for
	 * getrandom() does not work, because there's been no declaration for
	 * it for years.
	 */
	int res = getentropy_getrandom(buffer, length);

	if( res == -1 && errno == ENOSYS )
	{
		return getentropy_urandom(buffer, length);
	}

	return res;
}

static void
arc4random_buf (void *auth, int len)
{
    int	    ret;

#if HAVE_GETENTROPY
    /* weak emulation of arc4random through the getentropy libc call */
    ret = getentropy (auth, len);
#else
    ret = getentropy_emulate (auth, len);
#endif /* HAVE_GETENTROPY */
    if (ret == 0)
	return;

    insecure_getrandom_buf (auth, len);
}

#endif /* !defined(HAVE_ARC4RANDOM_BUF) */

void
XdmcpGenerateKey (XdmAuthKeyPtr key)
{
    arc4random_buf(key->data, 8);
}

int
XdmcpCompareKeys (const XdmAuthKeyPtr a, const XdmAuthKeyPtr b)
{
    int	i;

    for (i = 0; i < 8; i++)
	if (a->data[i] != b->data[i])
	    return FALSE;
    return TRUE;
}

void
XdmcpIncrementKey (XdmAuthKeyPtr key)
{
    int	i;

    i = 7;
    while (++key->data[i] == 0)
	if (--i < 0)
	    break;
}

void
XdmcpDecrementKey (XdmAuthKeyPtr key)
{
    int	i;

    i = 7;
    while (key->data[i]-- == 0)
	if (--i < 0)
	    break;
}
